import re

escapes = [
    (r"\\ ", "BACKSLASHSPACE"),
    (r"\\", "BACKSLASH"),
    ("~", "TILDE"),
    ("&", "\\&"),
    ("%", "\\%"),
    ("\\$", "\\$"),
    ("#", "\\#"),
    ("_", "\\_"),
    ("\\{", "\\{"),
    ("\\}", "\\}"),
    ("\\^", "\\^"),
    ("BACKSLASHSPACE", r"{\\textbackslash}~"),
    ("BACKSLASH", r"{\\textbackslash}"),
    ("TILDE", r"{\\textasciitilde}")
]

def escape_latex(content:str):
    if not content:
        return content
    for old, new in escapes:
        content = re.sub(old, new, content)
    return content



import typer
from ..client import Session


def get_me(
    token: str = typer.Argument(
        ..., envvar="CURVENOTE_TOKEN", help="API token generated from curvenote.com"
    )
):
    session = Session(token)
    typer.echo(session.user().json(indent=4))



import logging
import re

class SecureLogFormatter(logging.Formatter):
    def format(self, record):
        self._patterns = []
        msg = super(SecureLogFormatter, self).format(record)
        msg = re.sub(r'[A-Za-z0-9-_]*\.[A-Za-z0-9-_]*\.[A-Za-z0-9-_]', "***REMOVED***", msg)
        return msg

    def __getattr__(self, attr):
        return getattr(self.orig_formatter, attr)



from .StringSchemaOption import StringSchemaOption
from .BooleanSchemaOption import BooleanSchemaOption
from .SchemaOptionDefs import SchemaOptionDefs

from .AsideSchemaOption import AsideSchemaOption
from .CalloutSchemaOption import CalloutSchemaOption
from .CodeSchemaOption import CodeSchemaOption
from .HideLinksSchemaOption import HideLinksSchemaOption
from .RaggedBottomOption import RaggedBottomSchemaOption
from .SloppySchemaOption import SloppySchemaOption
from .CitepSchemaOption import CitepSchemaOption



PERMISSIVE_BLOCK_REGEX = r"block:[^\}]{6,}"
BLOCK_REGEX = r"block:[A-Za-z0-9]{3,}/[A-Za-z0-9]{3,}/[0-9]+"
UNPINNED_BLOCK_REGEX_ENDLINE = r"block:[A-Za-z0-9]{3,}/[A-Za-z0-9]{3,}$"
UNPINNED_BLOCK_REGEX_INLINE = r"block:[A-Za-z0-9]{3,}/[A-Za-z0-9]{3,}(?=[\s,}])"
GROUPED_BLOCK_REGEX = r"block:([A-Za-z0-9]{3,})/([A-Za-z0-9]{3,})/*([0-9]*)"

PERMISSIVE_OXA_REGEX = r"oxa:[^\}]{3,}"
OXA_REGEX = r"oxa:[A-Za-z0-9]{3,}/[A-Za-z0-9]{3,}\.[0-9]+"
UNPINNED_OXA_REGEX_ENDLINE = r"oxa:[A-Za-z0-9]{3,}/[A-Za-z0-9]{3,}$"
UNPINNED_OXA_REGEX_INLINE = r"oxa:[A-Za-z0-9]{3,}/[A-Za-z0-9]{3,}(?=[\s,}])"

# produces 2 or 3 groups
P1B2V_GROUPED_OXA_REGEX = r"oxa:([A-Za-z0-9]{3,})/([A-Za-z0-9]{3,})\.*([0-9]*)"

# produces 3 or 5 groups
P1B1VB2V_GROUPED_OXA_REGEX = r"oxa:([A-Za-z0-9]{3,})/([A-Za-z0-9]{3,})\.([0-9]+):([A-Za-z0-9]{3,})\.([0-9]+)[#0-9a-zA-Z]*$"

# produces 4 or 6 groups
# P1B1VP2B2V_GROUPED_OXA_REGEX = r"oxa:([A-Za-z0-9]{3,})/([A-Za-z0-9]{3,})\.*([0-9]*):([A-Za-z0-9]{3,})\.*([0-9]*)"

GROUPED_OXA_REGEX = P1B2V_GROUPED_OXA_REGEX

EXTRACT_UNPINNED_OXA_REGEX = r"(.*oxa:.{6,})\.[0-9#a-zA-Z]+"

CAPTION_BLOCK_REGEX = rf"{PERMISSIVE_BLOCK_REGEX}\.caption"
CAPTION_OXA_REGEX = rf"{PERMISSIVE_OXA_REGEX}\.caption"

INLINE_IMAGE_BLOCK_REGEX = (
    r".*\\includegraphics.*{(" + PERMISSIVE_BLOCK_REGEX + r")}.*\\*"
)
INLINE_IMAGE_OXA_REGEX = r".*\\includegraphics.*{(" + PERMISSIVE_OXA_REGEX + r")}.*\\*"

CAPTION_COMMAND_BLOCK_REGEX = r".*\\caption{(" + CAPTION_BLOCK_REGEX + r")}.*\\*"
CAPTION_COMMAND_OXA_REGEX = r".*\\caption{(" + CAPTION_OXA_REGEX + r")}.*\\*"
LABEL_COMMAND_REGEX = r".*\\label{([A-Za-z0-9_-]*)}.*\\*"
HREF_COMMAND_OXA_REGEX = r"\\href{(" + PERMISSIVE_OXA_REGEX + r")}"
REF_COMMAND_OXA_REGEX = r"\\ref{(" + PERMISSIVE_OXA_REGEX + r")}"

OUTPUT_IMAGE_BLOCK_REGEX = (
    r".*\\includegraphics.*{(" + PERMISSIVE_BLOCK_REGEX + r"-output-[0-9]+)}.*\\*"
)
OUTPUT_IMAGE_OXA_REGEX = (
    r".*\\includegraphics.*{(" + PERMISSIVE_OXA_REGEX + r"-output-[0-9]+)}.*\\*"
)

OUTPUT_SVG_BLOCK_REGEX = (
    r".*\\includesvg.*{(" + PERMISSIVE_BLOCK_REGEX + r"-output-[0-9]+)}.*\\*"
)
OUTPUT_SVG_OXA_REGEX = (
    r".*\\includesvg.*{(" + PERMISSIVE_OXA_REGEX + r"-output-[0-9]+)}.*\\*"
)

INLINE_CITATION_BLOCK_REGEX = r"\\cite[pt]?{(" + PERMISSIVE_BLOCK_REGEX + r")}"
INLINE_CITATION_OXA_REGEX = r"\\cite[pt]?{(" + PERMISSIVE_OXA_REGEX + r")}"

## FULL URLS
OXA_LINK = r"oxa:([a-zA-Z0-9]+)/([a-zA-Z0-9]+)\.([0-9]+)"
FULL_URL = r"http[s]*://.*/@([a-z0-9]{1}[a-z0-9_]+)/([a-zA-Z0-9]{1}[a-zA-Z0-9-]+)/([!a-zA-Z0-9]{1}[a-zA-Z0-9-]+)\.([0-9]+)"
PINNED_BLOCK_URL = r"http[s]*://.*/@([a-z0-9]{1}[a-z0-9_]+)/([a-zA-Z0-9]{1}[a-zA-Z0-9-]+)/([!a-zA-Z0-9]{1}[a-zA-Z0-9-]+)"
PROJECT_URL = r"http[s]*://.*/@([a-z0-9]{1}[a-z0-9_]+)/([a-zA-Z0-9-]+)"
TEAM_URL = r"http[s]*://.*/@([a-z0-9]{1}[a-z0-9_]+)"



import logging
import re
import time
from collections import namedtuple
from typing import List, NamedTuple, Optional, Type, Union

from pydantic.main import BaseModel

from ...client import Session
from ...models import Block, BlockFormat, BlockVersion
from .regex import INLINE_CITATION_BLOCK_REGEX
import requests
from collections import namedtuple
from typing import NamedTuple
from ...models import BlockVersion


class LocalMarker(NamedTuple):
    marker: str
    local_path: str
    remote_path: str
    content: Optional[str]

    """
        defining __eq__ and __hash__ is to allow the use of set()
    """
    def __eq__(self, other):
        return self.marker==other.marker

    def __hash__(self):
        return hash(('marker', self.marker))

# TODO move to session - easier to mock
def get_model(session: Session, url: str, model: Type[BaseModel] = BlockVersion):
    block = session.get_model(url, model)
    if not block:
        raise ValueError(f"Could not fetch the block {url}")
    return block


def fetch(url: str):
    resp = requests.get(url)
    if resp.status_code >= 400:
        raise ValueError(resp.content)
    return resp.content



from curvenote.models import BlockFormat
from curvenote.latex.utils.get_fast_hash import get_fast_hash
from typing import List
from .index import LocalMarker
from ...client import Session
from .regex import INLINE_IMAGE_BLOCK_REGEX, INLINE_IMAGE_OXA_REGEX
from .patch_local_images_from_tex_block import patch_local_images_from_tex_block
from .get_image_block import get_image_block


def localize_images_from_content_block(
    session: Session, assets_folder: str, figure_list: List[LocalMarker], content: str
):
    """Process image blocks and their respective referencing data including:
    - captions
    - labels
    """
    patch = patch_local_images_from_tex_block(
        [INLINE_IMAGE_BLOCK_REGEX, INLINE_IMAGE_OXA_REGEX], content
    )
    content_with_extensions = patch.content

    for remote_path, local_path in zip(patch.remote_paths, patch.local_paths):
        # get block
        image_block, local_path_with_extension = get_image_block(
            session, assets_folder, remote_path, local_path, BlockFormat.tex
        )

        # now patch the image captions which will be marked have placeholders like
        # f"{local_path}.caption"
        content_with_extensions = content_with_extensions.replace(
            f"{local_path}.caption", image_block.caption
        )

        # now patch the image labels which will have local_path based placeholders
        # like r"\label{local_path}" with a unique local marker
        figure = LocalMarker(
            marker=get_fast_hash(),
            local_path=local_path,
            remote_path=remote_path,
            content="",
        )
        figure_list.append(figure)

        content_with_extensions = content_with_extensions.replace(
            f"\\label{{{figure.local_path}}}", f"\\label{{{figure.marker}}}"
        )

        # patch the remaining image block references
        content_with_extensions = content_with_extensions.replace(
            local_path, local_path_with_extension
        )

    return content_with_extensions



import logging
import sys
import typer
import os
from curvenote.utils import decode_oxa_link, decode_url
from ..client import Session
from ..latex import LatexProjectBuilder, TemplateLoader


app = typer.Typer()

logger = logging.getLogger()


def url_set(ctx: typer.Context, value: str):
    if ctx.resilient_parsing:
        return
    if value:
        typer.echo("URL provided project, article and version options will be ignored")
    return value


def pull_as_latex(
    target: str = typer.Argument(
        ...,
        help=(
            "Local folder in which to construct the Latex assets. If TARGET exists it"
            "and all files will be removed and a new empty folder structure created"
        ),
    ),
    token: str = typer.Argument(
        ..., envvar="CURVENOTE_TOKEN", help="API token generated from curvenote.com"
    ),
    url: str = typer.Option(
        None,
        help=("A valid url (or oxa link) to the existing Article"),
        is_eager=True,
        callback=url_set,
    ),
    project: str = typer.Option(
        None,
        help=(
            "Identifier of existing Project containing ARTICLE. PROJECT may match title,"
            " name, or id of an existing Project. If no existing Project is found, an "
            "error will be raised"
        ),
    ),
    article: str = typer.Option(
        None,
        help=(
            "Identifier of existing Article. ARTICLE may match title, name, or id "
            "of an existing Article. If no existing Article is found, an error will"
            "be raised."
        ),
    ),
    version: int = typer.Option(
        None,  # optional
        help=(
            "Version of the article to pull, if not specified will pull the latest version."
        ),
    ),
    template: str = typer.Option(
        None,
        help=("Id (name) of the LaTeX template to apply"),
    ),
    local_template: str = typer.Option(
        None,
        help=("Local path to a folder containing a Curvenote compatible LaTeX template"),
    ),
    user_options: str = typer.Option(
        None,
        help=(
            "A path to a local YAML file containing user options to apply to the tempalte."
        ),
    ),
):
    logger.setLevel(logging.INFO)
    handler = logging.StreamHandler(sys.stdout)
    handler.setLevel(logging.INFO)
    formatter = logging.Formatter(
        "%(asctime)s - %(name)s - %(levelname)s - %(message)s"
    )
    handler.setFormatter(formatter)
    logger.addHandler(handler)

    if not url:
        if not project or not article:
            typer.echo("Either a --url or a --project and --article must be specified")
            raise typer.Exit(code=1)

    if user_options is not None and not os.path.exists(user_options):
        typer.echo(f"User options file note found on path {user_options}")
        raise typer.Exit(code=1)

    try:
        session = Session(token)
    except Exception as err:
        typer.echo("Could not start an API session - check your token")
        raise typer.Exit(code=1)

    loader = TemplateLoader(target)
    if template is not None:
        typer.echo(f"Using template: {template}")
        loader.initialise_from_template(session, template)
    elif local_template is not None:
        typer.echo(f"Using local template: {local_template}")
        loader.initialise_from_path(local_template)
    else:
        typer.echo(f"Using default template")
        loader.initialise_with_default()

    if user_options is not None:
        loader.set_user_options_from_yml(user_options)

    if url:
        typer.echo(f"Accessing block {url}")
        version_id, pathspec = None, None
        try:
            version_id = decode_oxa_link(url)
        except ValueError:
            pathspec = decode_url(url)
            typer.echo("Decoded {pathspec}")

        try:
            block = session.get_block(
                version_id.project if version_id is not None else pathspec.project,
                version_id.block if version_id is not None else pathspec.block,
            )
            typer.echo(f"Found & retreived block: {block.name}")
        except ValueError as err:
            typer.echo(err)
            typer.echo("Could not find block or you do not have access")
            raise typer.Exit(code=1) from err

        try:
            LatexProjectBuilder.build_single_article_by_url(loader, session, url)
        except ValueError as err:
            typer.echo(err)
            raise typer.Exit(code=1)

        typer.echo("Exiting...")
        return typer.Exit(code=0)

    try:
        project_obj = session.get_project(project)
        typer.echo(f"Found project: {project_obj.name}")
    except ValueError as err:
        typer.echo(f"Could not find project: {project} or you do not have access")
        raise typer.Exit(code=1) from err

    try:
        LatexProjectBuilder.build_single_article_by_name(
            loader, session, project_obj, article, version
        )
    except ValueError as err:
        typer.echo(err)
        raise typer.Exit(code=1)

    typer.Exit(code=0)



import re
from itertools import chain
from typing import List
from .localize_images import ImageSummary
from .run_regex_matchers import run_regex_matchers


def patch_local_images_from_tex_block(
    regex_matchers: List[str], content: str
) -> ImageSummary:
    """
    Take a piece of tex content that we expect to contain only one includegraphics directive
    get the remote path, assign a new local path and retrun a structure that can be parsed
    """
    updated_content = content
    matches = run_regex_matchers(regex_matchers, content)

    remote_paths: List[str] = []
    local_paths: List[str] = []
    for match in matches:
        remote_path = match[1]
        hash_segment = remote_path[6:]
        local_segment = hash_segment.replace("/", "_")
        local_path = (
            remote_path.replace("block:", "images/")
            .replace("oxa:", "images/")
            .replace(hash_segment, local_segment)
        )
        updated_content = updated_content.replace(remote_path, local_path)
        remote_paths.append(remote_path)
        local_paths.append(local_path)

    return ImageSummary(
        content=updated_content,
        remote_paths=remote_paths,
        local_paths=local_paths,
    )



import re
import logging
from .get_fast_hash import get_fast_hash


BIBTEX_ENTRY_TYPES = [
    "article",
    "book",
    "booklet",
    "conference",
    "inbook",
    "incollection",
    "inproceedings",
    "manual",
    "mastersthesis",
    "misc",
    "phdthesis",
    "proceedings",
    "techreport",
    "unpublished",
]


def parse_cite_tag_from_bibtex(content: str):
    logging.debug("bibtex: %s", content)
    tag = "ref"
    logging.debug(r"@[a-z]{3,}{([0-9a-zA-Z_]+),")
    match = re.match(r"@[a-z]{3,}{([0-9a-zA-Z_]+),", content)
    if match is not None:
        logging.debug("match %s", match)
        tag = match[1].replace(",", "")
    # adding quasi-random hash to save de-duping work
    hash_id = get_fast_hash()
    tag_with_hash = f"{tag}_{hash_id}"

    return tag_with_hash, tag



import re
from typing import List
from itertools import chain


def run_regex_matchers(regex_matchers: List[str], content: str):
    """
    Will run one of more regex matchers and concatenate the resulting
    match iterators
    """
    list_of_matches = []
    for matcher in regex_matchers:
        match = re.finditer(matcher, content)
        list_of_matches.append(match)
    return list(chain(*list_of_matches)) if len(list_of_matches) > 1 else list_of_matches[0]



import typer

from ..client import Session


def get_my_projects(
    token: str = typer.Argument(
        ..., envvar="CURVENOTE_TOKEN", help="API token generated from curvenote.com"
    )
):
    session = Session(token)
    for project in session.my_projects():
        typer.echo(project.json(indent=4))



import logging
import sys
import typer
from curvenote.models import (
    ArticleVersionPost,
    BlockChild,
    BlockKind,
    BlockPost,
    ContentVersionPost,
)

from ..client import Session

logger = logging.getLogger()


def clone_content(
    project: str = typer.Argument(
        ...,
        help=(
            "Identifier of existing Project containing ARTICLE. PROJECT may match title,"
            " name, or id of an existing Project. If no existing Project is found, an "
            "error will be raised"
        ),
    ),
    article: str = typer.Argument(
        ...,
        help=(
            "Identifier of existing Article. ARTICLE may match title, name, or id "
            "of an existing Article. If no existing Article is found, an error will"
            "be raised."
        ),
    ),
    new_article_name: str = typer.Argument(
        ...,
        help=(
            "Title of the new article to be created with cloned content."
            "If an article with the same name is found, an error will be raised."
        ),
    ),
    token: str = typer.Argument(
        ..., envvar="CURVENOTE_TOKEN", help="API token generated from curvenote.com"
    ),
    version: int = typer.Option(
        None,
        help=(
            "Version of the article to pull, if not specified will pull the latest version."
        ),
    ),
):
    logger.setLevel(logging.INFO)
    handler = logging.StreamHandler(sys.stdout)
    handler.setLevel(logging.INFO)
    formatter = logging.Formatter(
        "%(asctime)s - %(name)s - %(levelname)s - %(message)s"
    )
    handler.setFormatter(formatter)
    logger.addHandler(handler)

    session = Session(token)

    typer.echo("Checking for project access")
    try:
        project_obj = session.get_project(project)
    except ValueError as err:
        typer.echo(f"Unable to access project {project}")
        typer.Exit(code=1)

    typer.echo("Fetching existing article block")
    try:
        block = session.get_block(project_obj, article, BlockKind.article)
    except ValueError as err:
        typer.echo(f"Unable to access article {article}")
        typer.Exit(code=1)

    version = int(version) if version is not None else block.latest_version
    typer.echo("Fetching existing article version")
    try:
        version = session.get_version(block, version)
    except ValueError as err:
        typer.echo(f"Unable to get article version {version}")
        typer.echo(str(err))
        typer.Exit(code=1)

    new_order = []
    new_children = {}

    for o in version.order:
        child = version.children[o]
        try:
            child_block = session.get_block(project_obj, child.src.block)
        except ValueError as err:
            typer.echo(f"Unable to access child, skipping {child.src}")
            continue

        if child_block.kind != BlockKind.content:
            typer.echo(f"Found {child_block.kind} block, referencing")
            new_order.append(child.id)
            new_children[child.id] = child
        else:
            typer.echo("Found content block, cloning")
            child_version = session.get_version(child_block, child.src.version)
            new_child_block = session.upload_block(
                BlockPost(kind=child_block.kind), project=project_obj
            )
            new_child_version = session.upload_version(
                version=ContentVersionPost(content=child_version.content),
                block=new_child_block,
            )
            new_order.append(child.id)
            new_children[child.id] = BlockChild(id=child.id, src=new_child_version.id)

    typer.echo(f"Existing article had {len(version.order)} children")
    typer.echo(f"Cloned article will have {len(new_order)} children")

    typer.echo(f"Creating new article")
    try:
        new_article_block = session.upload_block(
            BlockPost(name=new_article_name, kind=BlockKind.article),
            project=project_obj,
        )
    except ValueError as err:
        typer.echo("Could not create new article block")
        typer.echo(str(err))
        typer.Exit(code=1)

    try:
        session.upload_version(
            version=ArticleVersionPost(
                order=new_order,
                children=new_children,
                title=new_article_name,
            ),
            block=new_article_block,
        )
    except ValueError as err:
        typer.echo("Could not create new article version")
        typer.echo(str(err))
        typer.Exit(code=1)

    typer.echo(f"New article created: {new_article_name}")



import logging
from curvenote.latex.utils.index import LocalMarker
from .utils.regex import HREF_COMMAND_OXA_REGEX
from typing import List
from curvenote.latex import utils

from ..client import Session
from ..models import (
    BlockKind,
    BlockVersion,
    VersionID,
)
from . import utils

def skipping_msg(thing: str, version_id: VersionID):
    return (
        f"Caught error trying to localize {thing} for block {str(version_id)}, skipping"
    )


class LatexBlockVersion:
    """
    Class to represent an block in the latex project.
    Encapulates operations applied in processing a block's content including
    calls to fetch content that it encapsulates

    Takes a reference to the reference_list and figure_list and updates these as a side effect
    """

    def __init__(
        self,
        session: Session,
        assets_folder: str,
        reference_list: List[LocalMarker],
        figure_list: List[LocalMarker],
        version: BlockVersion,
    ):
        self.session = session
        self.assets_folder = assets_folder
        self.reference_list = reference_list
        self.figure_list = figure_list
        self.version = version
        self._content = ""

    @property
    def content(self):
        return self._content

    def _localize_images_in_content(self, content: str):
        try:
            return utils.localize_images_from_content_block(
                self.session, self.assets_folder, self.figure_list, content
            )
        except Exception as err:  # pylint: disable=broad-except
            logging.warning(skipping_msg("image", self.version.id))
            logging.warning(str(err))
            return content

    def _localize_references_in_content(self, content: str):
        try:
            return utils.localize_references_from_content_block(
                self.session, self.reference_list, content
            )
        except Exception as err:  # pylint: disable=broad-except
            logging.warning(skipping_msg("reference", self.version.id))
            logging.warning(str(err))
            return content

    def _localize_href_in_content(self, content: str):
        try:
            return utils.localize_hrefs_in_content(self.session.site_url, content)
        except Exception as err:  # pylint: disable=broad-except
            logging.warning(skipping_msg("href", self.version.id))
            logging.warning(str(err))
            return content

    def localize(self):
        content = ""
        kind = self.version.kind
        if kind == BlockKind.content:
            logging.info("Found: Content Block")
            content = self.version.content
            content = self._localize_images_in_content(content)
            content = self._localize_references_in_content(content)
            content = self._localize_href_in_content(content)

        elif kind == BlockKind.output:
            logging.info(
                "Found: Output Block - num outputs: %s", len(self.version.outputs)
            )
            content = utils.localize_images_from_output_block(
                self.assets_folder, self.version
            )
        elif kind == BlockKind.code:
            logging.info("Found: Code Block")
            content = self.version.content
        elif kind == BlockKind.image:
            logging.info("Found: Top Level Image Block")
            content = utils.localize_image_from_top_level_block(
                self.session, self.assets_folder, self.figure_list, self.version
            )
        else:
            logging.warning("Can't process block with kind: %s yet", kind)
            raise ValueError(f"Unknown block kind {kind}")
        self._content = content



import logging
from typing import List
from curvenote.latex.utils.index import LocalMarker
from curvenote.latex.utils.links import version_id_to_local_path, version_id_to_oxa_path
from .get_image_block import get_image_block
from ...models import BlockFormat, BlockVersion
from ...client import Session
from .get_fast_hash import get_fast_hash


VERSION_ID = "___VERSION_ID___"
CAPTION = "___CAPTION___"
LABEL = "___LABEL___"
IMAGE_LATEX_SNIPPET = rf"""\begin{{figure}}[ht]
  \centering
  \includegraphics[width=1.0\textwidth]{{{VERSION_ID}}}
  \caption{{{CAPTION}}}
  \label{{{LABEL}}}
\end{{figure}}
"""


def localize_image_from_top_level_block(
    session: Session,
    assets_folder: str,
    figure_list: List[LocalMarker],
    version: BlockVersion,
):
    """
    - get the image block
    - download the image to the assets_folder using teh local_path as name
    - build the LaTeX content snippet and return it
    """
    try:
        oxa_path = version_id_to_oxa_path(version.id)
        local_path = version_id_to_local_path(version.id)

        image_block, local_path_with_extension = get_image_block(
            session, assets_folder, oxa_path, local_path, BlockFormat.tex
        )

        figure = LocalMarker(
            marker=get_fast_hash(),
            local_path=local_path,
            remote_path=oxa_path,
            content="",
        )
        figure_list.append(figure)

        content = (
            IMAGE_LATEX_SNIPPET.replace(VERSION_ID, local_path_with_extension)
            .replace(CAPTION, image_block.caption)
            .replace(LABEL, figure.marker)
        )

        return f"\n\n{content}\n"
    except ValueError as err:
        logging.error(
            "Caught error trying to localize top level image %s, skipping",
            str(version.id),
        )
        logging.error(err)



import os
import logging
import shutil
from .TemplateLoader import TemplateLoader
from curvenote.latex.utils.decorators import log_and_raise_errors
from typing import Union, List, NamedTuple
from pydantic import BaseModel
from curvenote.models import BlockFormat, Project
from ..client import Session
from ..utils import decode_url, decode_oxa_link
from .LatexArticle import LatexArticle
from .utils import LocalMarker, escape_latex


logger = logging.getLogger()


class ProjectItem(NamedTuple):
    path: str
    filename: str
    item: LatexArticle


class TemplateOptions(BaseModel):
    compact: bool = True


class LatexProjectBuilder:
    def __init__(self, loader: TemplateLoader, session: Session, target_folder: str):
        self.loader = loader
        self.session = session

        self.target_folder = os.path.abspath(target_folder)
        self.assets_folder = os.path.join(self.target_folder, "assets")
        self.images_folder = os.path.join(self.target_folder, "assets", "images")
        self.documents_folder = os.path.join(self.target_folder, "documents")

        logger.info("Creating %s", self.images_folder)
        os.makedirs(self.images_folder, exist_ok=True)
        logger.info("Creating %s", self.documents_folder)
        os.makedirs(self.documents_folder, exist_ok=True)

        self.articles: List[ProjectItem] = []
        self.reference_list: List[LocalMarker] = []
        self.figure_list: List[LocalMarker] = []

    @classmethod
    def build_single_article_by_name(
        cls,
        loader: TemplateLoader,
        session: Session,
        project_id_or_obj: Union[str, Project],
        article_id: str,
        version: int,
    ):
        latex_project = cls(loader, session, loader.target_folder)
        latex_project.add_article(
            project_id_or_obj, article_id, version, loader.options.tex_format
        )
        latex_project.reconcile()
        latex_project.write()

    @classmethod
    def build_single_article_by_url(
        cls, loader: TemplateLoader, session: Session, url: str
    ):
        vid, pathspec = None, None
        try:
            vid = decode_oxa_link(url)
        except ValueError:
            pathspec = decode_url(url)

        latex_project = cls(loader, session, loader.target_folder)

        logging.info("writing to %s", {latex_project.target_folder})
        latex_project.create_folders(loader.options.compact)

        if vid:
            latex_project.add_article(
                vid.project, vid.block, vid.version, loader.options.tex_format
            )
        else:
            if not pathspec.block:
                raise ValueError("URL does not include a block id")
            latex_project.add_article(
                pathspec.project,
                pathspec.block,
                pathspec.version,
                loader.options.tex_format,
            )

        latex_project.reconcile()
        latex_project.write()

    def create_folders(self, compact: bool = True):
        logger.info("Creating %s", self.images_folder)
        os.makedirs(self.images_folder, exist_ok=True)

        if not compact:
            logger.info("Creating %s", self.documents_folder)
            os.makedirs(self.documents_folder, exist_ok=True)

    def next_index(self):
        return len(self.articles)

    @log_and_raise_errors(lambda *args: "Could not add article to LaTeX project")
    def add_article(
        self,
        project_id: Union[str, Project],
        article_id: str,
        version: int,
        fmt: BlockFormat,
    ):
        logging.info("adding article using ids/names")
        latex_article = LatexArticle(self.session, project_id, article_id)
        latex_article.fetch(fmt, version)
        latex_article.localize(
            self.session, self.assets_folder, self.reference_list, self.figure_list
        )
        filename = f"{self.next_index()}_{latex_article.block.name}"
        self.articles.append(
            ProjectItem(
                path=f"documents/{filename}", item=latex_article, filename=filename
            )
        )
        logging.info("added article")

    def reconcile(self):
        for article in self.articles:
            article.item.reconcile_figures(self.figure_list)

    def render(self):
        logging.info("Rendering template...")
        if self.loader.renderer is None:
            logging.info("TemplateRender is not available, TemplateLoader not initialized")
            raise ValueError("TemplateRender is not available, TemplateLoader not initialized")

        if len(self.articles) < 1:
            raise ValueError("Need at least one article")

        content = ""
        for article in self.articles:
            content += article.item.content

        first = self.articles[0]

        authors = []
        for author in first.item.authors:
            if author.user:
                user = next((u for u in first.item.users if u.id == author.user), None)
                if user:
                    authors.append(dict(
                        username = user.username,
                        name = user.display_name,
                        bio = user.bio,
                        location = user.location,
                        website = user.website,
                        github = user.github,
                        twitter = user.twitter,
                        affiliation = user.affiliation,
                        orcid = user.orcid,
                    ))
            elif author.plain:
                authors.append(dict(name=escape_latex(author.plain)))
            else:
                logging.info("found empty author, skipping...")

        data = dict(
            doc=dict(
                oxalink=first.item.oxalink(self.session.site_url),
                title=escape_latex(first.item.title),
                authors=authors,
                date=first.item.date,
                tags=[escape_latex(tag) for tag in first.item.tags ]
            ),
            tagged=dict(),
            curvenote=dict(defs=r"\input{curvenote.def}"),
            options=dict(docclass=''),
            CONTENT=content,
        )

        return self.loader.renderer.render(data)

    @log_and_raise_errors(lambda *args: "Could not write final document")
    def write(self):
        logging.info("ProjectBuilder - writing...")

        if not self.loader.options.compact:
            for article in self.articles:
                article_filepath = os.path.join(
                    self.documents_folder, article.filename + ".tex"
                )
                article.item.write(article_filepath)

        content = self.render()

        logging.info("Building template defs")
        content_transforms = self.loader.build_defs()

        logging.info("Applying content transforms")
        for transform in content_transforms:
            content = transform(content)

        logging.info("Writing index.tex...")
        with open(os.path.join(self.target_folder, "index.tex"), "w+") as file:
            file.write(content)

        logging.info("Writing main.bib...")
        if len(self.reference_list) > 0:
            # de-duplicate the bib entries
            bib_entries = list(set(self.reference_list))
            with open(os.path.join(self.target_folder, "main.bib"), "w+") as file:
                for reference in bib_entries:
                    file.write(f"{reference.content}\n")

        logging.info("Cleaning up...")
        files = ["template.tex", "template.yml"]
        for file in files:
            if os.path.exists(os.path.join(self.target_folder, file)):
                shutil.move(os.path.join(self.target_folder, file), os.path.join(self.target_folder, f"{file}.ignore"))

        logging.info("Done!")



import typer
from .version import __version__

app = typer.Typer()

def version_callback(value: bool):
    if value:
        typer.echo(
            r"""
                LaTeX Template Generating System by Curvenote
            """
        )
        typer.echo(f"Version: {__version__}")
        raise typer.Exit()

@app.callback()
def main(
    version: bool = typer.Option(
        None, "--version", callback=version_callback, is_eager=True
    )
):
    return



import re
import logging
from .SchemaOptionDefs import SchemaOptionDefs
from .BooleanSchemaOption import BooleanSchemaOption


def citep_transform(content: str):
    logging.info("Transform: citep_transform")
    return re.sub(r"\\citep{", r"\\cite{", content)

class CitepSchemaOption(BooleanSchemaOption):
    def __init__(self):
        super(CitepSchemaOption, self).__init__()

        self.add(True, SchemaOptionDefs(
            packages=["package-natbib.def"],
            setup=["setup-bib-style.def", "setup-cite-style.def"]
            )
        )

        self.add(False, SchemaOptionDefs(
            transforms=[citep_transform]
        ))

        self.set_default(True)



from .LatexArticle import LatexArticle
from .LatexProjectBuilder import LatexProjectBuilder
from .LatexBlockVersion import LatexBlockVersion
from .TemplateLoader import TemplateLoader
from .templating.TemplateOptions import TemplateOptions
from .templating.TemplateRenderer import TemplateRenderer



import re
from curvenote.latex.utils.regex import (
    FULL_URL,
    GROUPED_BLOCK_REGEX,
    GROUPED_OXA_REGEX,
    OXA_LINK,
    P1B1VB2V_GROUPED_OXA_REGEX,
    PINNED_BLOCK_URL,
    PROJECT_URL,
    TEAM_URL,
    EXTRACT_UNPINNED_OXA_REGEX,
)
from curvenote.models import BlockChildID, BlockFormat, PathSpec, VersionID
from typing import Union


def block_hash_to_api_url(
    api_url: str, block_hash: str, fmt: Union[BlockFormat, None] = None
):
    """
    Converts a block hash with format:
        `block:project1/block1/v

    into a valid api url
    """
    query = ""
    if fmt is not None:
        query = f"?format={fmt}"
    match = re.search(GROUPED_BLOCK_REGEX, block_hash)
    if match:
        if match[3]:
            return f"{api_url}/blocks/{match[1]}/{match[2]}/versions/{match[3]}{query}"
        return f"{api_url}/blocks/{match[1]}/{match[2]}{query}"
    raise ValueError(f"invalid block hash {block_hash}")


def oxa_path_to_api_url(
    api_url: str, oxa_path: str, fmt: Union[BlockFormat, None] = None
):
    """
    Converts a oxa path with format:
        `oxa:project1/block1.version

    into a valid api url
    """
    query = ""
    if fmt is not None:
        query = f"?format={fmt}"
    match = re.search(GROUPED_OXA_REGEX, oxa_path)
    if match:
        if match[3]:
            return f"{api_url}/blocks/{match[1]}/{match[2]}/versions/{match[3]}{query}"
        return f"{api_url}/blocks/{match[1]}/{match[2]}{query}"
    raise ValueError(f"invalid oxa path {oxa_path}")


def decode_oxa_link(url: str) -> VersionID:
    """
    Retreive a valid Curvenote VersionID from the oxa link
    """
    match = re.search(OXA_LINK, url)
    if match:
        project, block, version = match.groups()
        return VersionID(project=project, block=block, version=int(version))
    raise ValueError("Not a valid oxa link")


def decode_url(url: str) -> Union[VersionID, PathSpec]:
    """
    Retreive a valid Curvenote path spec from the url
    """
    match = re.search(FULL_URL, url)
    if match:
        team, project, block, version = match.groups()
        return PathSpec(
            team=team,
            project=project,
            block=block[1:] if block[0] == "!" else block,
            version=int(version),
        )

    match = re.search(PINNED_BLOCK_URL, url)
    if match:
        team, project, block = match.groups()
        return PathSpec(
            team=team,
            project=project,
            block=block[1:] if block[0] == "!" else block,
        )

    match = re.search(PROJECT_URL, url)
    if match:
        team, project = match.groups()
        return PathSpec(
            team=team,
            project=project,
        )

    match = re.search(TEAM_URL, url)
    if match:
        print(match.groups())
        (team,) = match.groups()
        return PathSpec(team=team)

    raise ValueError("Not a valid Curvenote url")


def version_id_to_block_path(block_id: BlockChildID):
    return f"block:{block_id.project}/{block_id.block}/{block_id.version}"


def version_id_to_oxa_path(block_id: BlockChildID):
    return f"oxa:{block_id.project}/{block_id.block}.{block_id.version}"


def version_id_to_local_path(block_id: BlockChildID):
    return f"images/{block_id.project}_{block_id.block}_{block_id.version}"


def unpin_oxa_path(oxa_path: str):
    """
    Take an oxa path or oxa link and return the unpinned version
    Should work for links as well as paths

    e.g.
        oxa:project1/block2.1 -> oxa:project1/block2
        oxa:project1/block1.1:project2/block2.1#id -> oxa:block1.1:project2/block2
        http://api.curvenote.com:8888/oxa:project1/block2.1 -> http://api.curvenote.com:8888/oxa:project1/block2

    If there is no match we assume it's already unpinned and return the argument
    """
    match = re.search(EXTRACT_UNPINNED_OXA_REGEX, oxa_path)
    return oxa_path if match is None else match[1]


def minimise_oxa_path(oxa_path: str):
    """
    Take an oxa path or oxa link and return the most compact version of it without any context

    e.g.
        oxa:project1/block1.1:project2/block2.1#id -> oxa:block1.1:project2/block2

    If no match for a long form oxa path is found, the argument will be returned

    TODO needs extended to other cases
    """
    match = re.search(P1B1VB2V_GROUPED_OXA_REGEX, oxa_path)
    if match is None:
        return oxa_path

    return f"oxa:{match[1]}/{match[4]}.{match[5]}"



import logging
import sys
import subprocess
import typer

app = typer.Typer()

logger = logging.getLogger()


def build_pdf(
    target: str = typer.Argument(
        ...,
        help=(
            "Local folder containing the local LaTeX project."
            "Must contain an index.tex file."
        ),
    ),
):
    logger.setLevel(logging.INFO)
    handler = logging.StreamHandler(sys.stdout)
    handler.setLevel(logging.INFO)
    formatter = logging.Formatter(
        "%(asctime)s - %(name)s - %(levelname)s - %(message)s"
    )
    handler.setFormatter(formatter)
    logger.addHandler(handler)

    typer.echo("Build PDF from LaTeX")
    typer.echo(f"Target folder: {target}")
    typer.echo("Invoking xelatex...")
    try:
        XELATEX_CMD = f'cd {target}; latexmk -f -xelatex -synctex=1 -interaction=nonstopmode -file-line-error -latexoption="-shell-escape" index.tex'
        ret_val = subprocess.run(XELATEX_CMD, shell=True)
        if ret_val.returncode != 0:
            typer.echo(
                f"mklatex returned with a non zero error code, but thePDF was still created"
            )
            typer.Exit(code=1)
    except ValueError as err:
        typer.echo(f"Fatal error while running mklatex")
        typer.Exit(code=1)

    typer.echo(f"mklatex reports success!")
    typer.Exit(code=0)



from collections import namedtuple

# TODO these should be somewhere else and an enums or pydantic models?
ImageFormats = {"image/png": "png", "image/jpeg": "jpg", "image/gif": "gif"}
ImageSummary = namedtuple("ImageSummary", ["content", "remote_paths", "local_paths"])



import re


def find_groups_in_content(regex_matcher: str, content: str):
    """
    Take a piece of tex content that we expect to contain one or more commands,
    find all instances based on the regex given and return the first groups

    Typically this is used to extract block_paths for different latex commands from
    TeX content
    """
    block_paths = []
    matches = re.finditer(regex_matcher, content)
    for match in matches:
        block_path = match[1]
        block_paths.append(block_path)

    return block_paths



from .links import oxa_path_to_api_url
from .regex import HREF_COMMAND_OXA_REGEX
from .run_regex_matchers import run_regex_matchers


def localize_hrefs_in_content(site_url: str, content: str):
    matches = run_regex_matchers([HREF_COMMAND_OXA_REGEX], content)
    for match in matches:
        url = oxa_path_to_api_url(site_url, match[1])
        content = content.replace(match[0], f"\\href{{{url}}}")
    return content



import logging
import os
from jinja2.loaders import PackageLoader
import yaml
import pkg_resources
from typing import Dict
from zipfile import ZipFile
from shutil import copyfile
from distutils.dir_util import copy_tree
from .templating.TemplateRenderer import TemplateRenderer
from .templating.TemplateOptions import TemplateOptions
from ..utils import download
from ..client import Session
from .DefBuilder import DefBuilder

TEMPLATES_PATH = pkg_resources.resource_filename("curvenote", "latex/templates")
DEFAULT_TEMPLATE_PATH = os.path.join(TEMPLATES_PATH, "default")

class TemplateLoader:
    def __init__(self, target_folder: str):
        self.template_name = None
        self.renderer = None
        self.options = None

        self.target_folder = target_folder
        os.makedirs(self.target_folder, exist_ok=True)

    def initialise_with_default(self):
        logging.info("TemplateLoader - Initialising with default template")
        logging.info("Copying template assets")
        template_location = pkg_resources.resource_filename(
            "curvenote", os.path.join("latex", "templates", "default")
        )
        template_assets = ["curvenote.png", "template.yml"]
        for asset_filename in template_assets:
            src = os.path.join(template_location, asset_filename)
            dest = os.path.join(self.target_folder, asset_filename)
            logging.info("Copying: %s to %s", src, dest)
            copyfile(src, dest)

        self.template_name = "default"
        self.renderer = TemplateRenderer()
        self.renderer.use_loader(PackageLoader("curvenote", os.path.join("latex", "templates", "default")))
        self.options = TemplateOptions(DEFAULT_TEMPLATE_PATH)

    def initialise_from_path(self, local_template_path: str):
        logging.info("Using local template found at: %s", local_template_path)

        abs_path = os.path.abspath(os.path.expanduser(local_template_path))

        if not os.path.exists(abs_path):
            raise ValueError('local template path does not exist')

        if not os.path.isdir(abs_path):
            raise ValueError('local template path must point to a folder')

        try:
            copy_tree(abs_path, self.target_folder)
        except Exception as err:
            logging.error("Could not copy local template from %s to %s", abs_path, self.target_folder)
            logging.error(str(err))
            raise err

        self.template_name = os.path.basename(os.path.normpath(abs_path))
        self.renderer = TemplateRenderer()
        self.renderer.use_from_folder(self.target_folder)
        self.options = TemplateOptions(self.target_folder)

    def initialise_from_template(self, session: Session, template_name: str):
        logging.info("Writing to target folder: %s", self.target_folder)

        logging.info("Looking up template %s", template_name)
        try:
            link = session.get_template_download_link(template_name)
        except ValueError as err:
            logging.error("could not download template %s", template_name)
            raise ValueError(f"could not download template: {template_name}") from err

        # fetch template to local folder
        logging.info("Found template, downloading...")
        zip_filename = os.path.join(self.target_folder, f"{template_name}.template.zip")
        download(link, zip_filename)

        # unzip
        logging.info("Download complete, unzipping...")
        with ZipFile(zip_filename, "r") as zip_file:
            zip_file.extractall(self.target_folder)
        logging.info("Unzipped to %s", self.target_folder)
        os.remove(zip_filename)
        logging.info("Removed %s", zip_filename)

        # success -- update members
        self.template_name = template_name
        self.renderer = TemplateRenderer()
        self.renderer.use_from_folder(self.target_folder)
        self.options = TemplateOptions(self.target_folder)

    def set_user_options_from_yml(self, user_options_path: str):
        if self.options is None:
            raise ValueError('TemplateLoader not initialized')

        with open(user_options_path, "r") as file:
            self.options.set_user_options(yaml.load(file, Loader=yaml.FullLoader))

    def set_user_options(self, user_options: Dict):
        if self.options is None:
            raise ValueError('TemplateLoader not initialized')
        self.options.set_user_options(user_options)

    def build_defs(self):
        """
        TODO: regsiter schema options here with:
            - template config path
            - a map of:
                - paths to definitions for known options
                - package dependencies for known options

        1. iterate over all *registered* schema options
            1.1 check if option is speced in the template, using the template setting or
                using the default
            1.2 log any package dependencied in the package listing
        """
        if self.template_name is None or self.options is None:
            raise ValueError("TemplateLoader has not been initialized")
        def_builder = DefBuilder()
        def_builder.build(self.options, self.target_folder)
        return def_builder.get_content_transforms(self.options)



from curvenote.models import BlockFormat
import os
import pkg_resources
import logging
import yaml
from typing import Dict, Any, Optional, Union
from pykwalify.core import Core as YamlSchema


SCHEMA_PATH = pkg_resources.resource_filename("curvenote", "latex/templates")

class TemplateOptions:
    _tex_format: BlockFormat

    def __init__(self, template_location: str):
        self.template_schema = None
        self.schemas = [
            os.path.join(SCHEMA_PATH, "config.schema.yml"),
            os.path.join(SCHEMA_PATH, "template.schema.yml"),
        ]

        self.template_location = template_location
        template_yml = os.path.join(template_location, "template.yml")
        print("Looking for template on %s", template_yml)
        if not os.path.exists(template_yml):
            logging.info("%s does not exist", template_yml)
            raise FileNotFoundError(f"{template_yml} does not exist")

        self.template_schema = YamlSchema(
            source_file=template_yml, schema_files=self.schemas
        )
        self.template_schema.validate(raise_exception=True)
        self.flat_config = TemplateOptions._flatter(
            self.template_schema.source["config"]
        )
        self._tex_format = (
            BlockFormat.tex
            if self.get("config.build.vanilla")
            else BlockFormat.tex_curvenote
        )

    @property
    def tex_format(self):
        return self._tex_format

    def __len__(self):
        if self.template_schema is None:
            return 0
        return len(self.flat_config)

    def get(self, path: str, default: Any = None):
        """
        Get a value from the template options on the specified path

        raises a ValueError if the options is not found
        """
        try:
            return TemplateOptions.find(path, self.template_schema.source)
        except KeyError:
            return default

    def dumps(self):
        """
        Dump the template options as a YAML string
        """
        return yaml.dump(self.template_schema.source)

    @staticmethod
    def _flatter(options):
        """
        single level flattening and path concatenation
        """
        flat = []
        for top, group in options.items():
            for k, value in group.items():
                flat.append((f"{top}.{k}", value))
        return flat

    @staticmethod
    def find(element: str, data: Dict):
        keys = element.split(".")
        rv = data
        for key in keys:
            if key not in rv:
                raise KeyError(f"{key} not found")
            rv = rv[key]
        return rv

    @staticmethod
    def templates_path():
        return SCHEMA_PATH

    def set_user_options(self, user_options: Dict):
        """
        Parse a dict of user options, validate these against the template and
        register any valid options.

        Discard any option not listed in or conforming to the template config.options section
        """

        # TODO register user defined options
        pass

    @property
    def compact(self):
        return self.get("config.build.layout") == "compact"

    @property
    def schema(self) -> Dict[str,Union[str, bool]]:
        """
        Return the schema secton
        """
        if self.template_schema is None or self.template_schema.source is None:
            return {}
        return self.template_schema.source["config"]["schema"]

    @property
    def tagged(self):
        """
        Return the tagged section
        """
        return (
            self.template_schema.source["config"]["tagged"]
            if self.template_schema
            else None
        )

    @property
    def user_options(self):
        """
        TODO
        """
        return (
            self.template_schema.source["config"]["options"]
            if self.template_schema
            else None
        )



import re
import logging
from typing import List
from curvenote.models import Block, BlockVersion
from .parse_cite_tag_from_bibtex import parse_cite_tag_from_bibtex
from .links import block_hash_to_api_url, oxa_path_to_api_url, unpin_oxa_path
from .index import (
    LocalMarker,
    get_model,
)
from ...client import Session
from .regex import (
    BLOCK_REGEX,
    INLINE_CITATION_BLOCK_REGEX,
    INLINE_CITATION_OXA_REGEX,
    OXA_REGEX,
    UNPINNED_BLOCK_REGEX_ENDLINE,
    UNPINNED_BLOCK_REGEX_INLINE,
    UNPINNED_OXA_REGEX_ENDLINE,
    UNPINNED_OXA_REGEX_INLINE,
)
from .run_regex_matchers import run_regex_matchers


def split_remote_paths(remote_path: str) -> List[str]:
    """Splits a remote path into a list of block_paths or oxa_paths"""
    matches = run_regex_matchers([
            OXA_REGEX, UNPINNED_OXA_REGEX_ENDLINE,
            UNPINNED_OXA_REGEX_INLINE, BLOCK_REGEX,
            UNPINNED_BLOCK_REGEX_ENDLINE, UNPINNED_BLOCK_REGEX_INLINE
            ], remote_path)

    if len(matches) == 0:
        return [remote_path]

    if len(matches) > 1:
        logging.info("Found group of %s paths, splitting...", len(matches))

    return [m[0] for m in matches]


def localize_references_from_content_block(
    session: Session, reference_list: List[LocalMarker], content: str
):
    """Looks for cite TeX commands in the content then replaces the block ids
    with locally unique identifiers based on the local reference list.

    The reference list is extended as new references are found (side effect)

    Appends a unique hash to each new reference encountered
    """
    patched_content = content

    matches = run_regex_matchers(
        [INLINE_CITATION_BLOCK_REGEX, INLINE_CITATION_OXA_REGEX], content
    )

    # citep blocks can contain multiple citations, so we need to iterate over
    # the list of matches, check each one and split them
    remote_paths: List[str] = []
    for match in matches:
        remote_paths += split_remote_paths(match[1])

    for remote_path in remote_paths:
        logging.info("processing remote path: %s", remote_path)
        # check for the reference in the reference list based on the block_path
        matched_references = [r for r in reference_list if r.remote_path == remote_path]
        existing_reference = (
            matched_references[0] if (len(matched_references) > 0) else None
        )

        if existing_reference is None:
            if remote_path.startswith("oxa:"):
                unpinned_path = unpin_oxa_path(remote_path)
                url = oxa_path_to_api_url(session.api_url, unpinned_path)
            else:
                url = block_hash_to_api_url(session.api_url, remote_path)

            logging.info("fetching reference block %s", url)
            block = get_model(session, url, Block)
            logging.info("got reference block")

            # get latest version
            version_url = f"{url}/versions/{block.latest_version}?format=bibtex"
            logging.info("fetching reference version %s", version_url)
            version = get_model(session, version_url, BlockVersion)
            logging.info("got reference version")

            # update the list
            bibtex_content: str = version.content
            _, plain_tag = parse_cite_tag_from_bibtex(bibtex_content)

            reference_item = LocalMarker(plain_tag, remote_path, plain_tag, bibtex_content)
            reference_list.append(reference_item)
            existing_reference = reference_item
            logging.info("using new reference %s", existing_reference.marker)
        else:
            logging.info("using existing reference %s", existing_reference.marker)

        # patch the content and move on
        patched_content = patched_content.replace(
            remote_path, existing_reference.marker
        )

    return patched_content



from .index import *
from .regex import *
from .imagemagick import *
from .links import *
from .patch_local_images_from_tex_block import patch_local_images_from_tex_block
from .localize_images_from_content_block import localize_images_from_content_block
from .localize_images_from_output_block import localize_images_from_output_block
from .localize_image_from_top_level_block import localize_image_from_top_level_block
from .localize_references_from_content_block import (
    localize_references_from_content_block,
)
from .localize_hrefs_in_content import localize_hrefs_in_content
from .get_fast_hash import get_fast_hash
from .parse_cite_tag_from_bibtex import parse_cite_tag_from_bibtex
from .decorators import just_log_errors, log_and_raise_errors
from .escape_latex import escape_latex



import os
import re
import requests
from sre_constants import GROUPREF
from typing import Dict, Union, cast
from urllib.parse import parse_qs, urlencode, urlparse, urlunparse
from .models import BlockID, PathSpec, VersionID, SrcID
from pydantic import AnyHttpUrl


def update_query(
    url: AnyHttpUrl,
    query: Dict[str, str],
) -> AnyHttpUrl:
    """Add query strings to URL"""
    parsed_url = urlparse(url)
    parsed_query = parse_qs(parsed_url.query)
    parsed_query.update(query)
    unparsed_url = urlunparse(
        tuple(parsed_url[:4] + (urlencode(parsed_query, doseq=True),) + parsed_url[5:])
    )
    return unparsed_url


def is_url(value: str) -> bool:
    """Returns True if value is URL"""
    return urlparse(value).scheme and urlparse(value).netloc


def github_to_raw(url: AnyHttpUrl) -> AnyHttpUrl:
    """Converts a github.com URL to raw.githubusercontent.com URL...

    Sketchy.
    """
    return url.replace("github.com", "raw.githubusercontent.com").replace("/blob/", "/")


def title_to_name(title: str) -> str:
    """Replace all non-alphanumeric characters with hyphens"""
    name = re.sub("[^0-9a-z]+", "-", title.lower())[:50]
    while name.startswith("-"):
        name = name[1:]
    while name.endswith("-"):
        name = name[:-1]
    return name


def filename_to_title(filename: str) -> str:
    """Remove path and extension from filename"""
    return filename.split(os.path.sep)[-1].rsplit(".", 1)[0]


OXA_LINK = r"oxa:([a-zA-Z0-9]+)/([a-zA-Z0-9]+)\.([0-9]+)"
FULL_URL = r"http[s]*://.*/@([a-z0-9]{1}[a-z0-9_]+)/([a-zA-Z0-9]{1}[a-zA-Z0-9-]+)/([!a-zA-Z0-9]{1}[a-zA-Z0-9-]+)\.([0-9]+)"
PINNED_BLOCK_URL = r"http[s]*://.*/@([a-z0-9]{1}[a-z0-9_]+)/([a-zA-Z0-9]{1}[a-zA-Z0-9-]+)/([!a-zA-Z0-9]{1}[a-zA-Z0-9-]+)"
PROJECT_URL = r"http[s]*://.*/@([a-z0-9]{1}[a-z0-9_]+)/([a-zA-Z0-9-]+)"
TEAM_URL = r"http[s]*://.*/@([a-z0-9]{1}[a-z0-9_]+)"


def decode_oxa_link(url: str) -> VersionID:
    """
    Retreive a valid Curvenote VersionID from the oxa link
    """
    match = re.search(OXA_LINK, url)
    if match:
        project, block, version = match.groups()
        return VersionID(project=project, block=block, version=int(version))
    raise ValueError("Not a valid oxa link")


def decode_url(url: str) -> Union[VersionID, PathSpec]:
    """
    Retreive a valid Curvenote path spec from the url
    """
    match = re.search(FULL_URL, url)
    if match:
        team, project, block, version = match.groups()
        return PathSpec(
            team=team,
            project=project,
            block=block[1:] if block[0] == "!" else block,
            version=int(version),
        )

    match = re.search(PINNED_BLOCK_URL, url)
    if match:
        team, project, block = match.groups()
        return PathSpec(
            team=team,
            project=project,
            block=block[1:] if block[0] == "!" else block,
        )

    match = re.search(PROJECT_URL, url)
    if match:
        team, project = match.groups()
        return PathSpec(
            team=team,
            project=project,
        )

    match = re.search(TEAM_URL, url)
    if match:
        print(match.groups())
        (team,) = match.groups()
        return PathSpec(team=team)

    raise ValueError("Not a valid Curvenote url")


def fetch(url: str):
    resp = requests.get(url)
    if resp.status_code >= 400:
        raise ValueError(resp.content)
    return resp.content


def download(url, save_path, chunk_size=128):
    """
    Download a file from a url and save to the save_path provided
    """
    resp = requests.get(url, stream=True)
    with open(save_path, "wb") as file:
        for chunk in resp.iter_content(chunk_size=chunk_size):
            file.write(chunk)

def id_to_api_url(base: str, the_id: Union[BlockID,VersionID,SrcID]):
    if 'version' in the_id.__dict__:
        version_id = cast(VersionID, the_id)
        return f"{base}/blocks/{version_id.project}/{version_id.block}/versions/{version_id.version}"
    else:
        return f"{base}/blocks/{the_id.project}/{the_id.block}"



import typer
from .get_me import get_me
from .get_my_projects import get_my_projects
from .push import push
from .clone_content import clone_content
from .pull_as_latex import pull_as_latex
from .build_pdf import build_pdf

from ..version import __version__

app = typer.Typer()
app.command()(get_me)
app.command()(get_my_projects)
app.command()(push)
app.command()(clone_content)
app.command()(pull_as_latex)
app.command()(build_pdf)


def version_callback(value: bool):
    if value:
        typer.echo(
            r"""
   ______                                  __
  / ____/_  ________   _____  ____  ____  / /____
 / /   / / / / ___/ | / / _ \/ __ \/ __ \/ __/ _ \
/ /___/ /_/ / /   | |/ /  __/ / / / /_/ / /_/  __/
\____/\__,_/_/    |___/\___/_/ /_/\____/\__/\___/
        """
        )
        typer.echo(f"Curvenote CLI Version: {__version__}")
        raise typer.Exit()


@app.callback()
def main(
    version: bool = typer.Option(
        None, "--version", callback=version_callback, is_eager=True
    )
):
    return
